import re
import ast
import codegen


def find(inString, findx):
    '''
    Find start locations of all instances of a substring within a string
    :param inString: String to search through
    :param findx: Substring to seach for
    :return: Start locations of all substring positions
    '''
    return [m.start() for m in re.finditer(findx, inString)]


def escape_forbidden_characters(in_string):
    '''
    Escape forbidden characters in input string in accordance with PROV-N
    :param in_string: input string to escape characters within
    :return: modified in_string with escaped characters
    '''
    forbidden = "\\(=)|,-:;[]+"
    allowed_special = '/._'
    in_string = [x if x.isalnum() else '\\'+x if x in forbidden else x if x in allowed_special else '' for x in in_string]
    final_string = ''.join(in_string)
    return final_string


def separateLists(inList):
    '''
    Similar to a ','.split(inList) function, but it also corrects for peculiarities in python parameter formatting
    :param inList: String representation of a list to separate
    :return: True list of separated variable names
    '''
    findC = ','
    sepLocs = find(inList,findC)
    sepLocs.insert(0,0)
    sepLocs.append(len(inList))
    
    inputList = [inList[x:y] for x, y in zip(sepLocs[0:-1], sepLocs[1:])]
    inputList = [x.replace(',','') for x in inputList]
    
    for x,inL in enumerate(inputList):
        if '=' in inL:
            loc1 = inL.find('=')
            inVal = inL[:loc1]
            inputList[x] = inVal
    
    return inputList


def separate_output(in_string):
    '''
    Walks the AST of a python function in order to separate outputs into distinct objects where applicable
    :param in_string: Function source code as string to walk
    :return: List of outputs from function
    '''
    answer = [(node.value) for node in ast.walk(ast.parse(in_string)) if isinstance(node, ast.Return)]

    possible_outputs = []
    for a in answer:
        try:
            outputlist = [codegen.to_source(a)]
        except (AttributeError, TypeError) as e:
            outputlist = ['None']
        if isinstance(a, tuple):
            outputlist = separateLists(outputlist)
        possible_outputs.append(outputlist)

    if len(possible_outputs) > 0:
        output_list = possible_outputs[-1]
    else:
        output_list = []

    return output_list


def variableSeparation(inString):
    '''
    Deconstructs function source code into key components such as name, inputs, and outputs
    :param inString: Function source code to separate
    :return: Dictionary of separated objects
    '''
    find1 = 'def '
    find2 = '('
    find3 = '):'
    find4 = 'return '
    find5 = '\n'
    
    functionDict = {}
    
    functionLocs = find(inString,find1)
    functionLocs.append(len(inString))

    for fNum in range(len(functionLocs)-1):
        searchString = inString[functionLocs[fNum]:functionLocs[fNum+1]]
        loc2 = searchString.index(find2)
        loc3 = searchString.index(find3)

        try:
            loc4s = find(searchString,find4)
        except:
            loc4s = []
        loc5s = [searchString[x:].index(find5) + x for x in loc4s]
        functName = searchString[len(find1):loc2]
        inputList = searchString[loc2+len(find2):loc3]
        if ',' in inputList:
            inputList = separateLists(inputList)
        else:
            inputList = [inputList]

        try:
            outputList = searchString[loc4s[-1]+len(find4):loc5s[-1]]
        except IndexError:
            outputList = 'empty'

        if ',' in outputList:
            outputList = separateLists(outputList)
        else:
            outputList = [outputList]

        if len(loc5s) != 0:
            function_string = searchString[:loc5s[-1]]
        else:
            function_string = searchString

        functionDict['funct_{0}'.format(fNum)] = [functName,inputList,outputList,function_string]
    return functionDict



bundleStart = """
document
prefix prov <http://www.w3.org/ns/prov#>
prefix var <http://openprovenance.org/var#>
prefix exe <http://example.org/>
prefix prtr <https://praetor.pages.mpcdf.de/praetor_provenance/>

bundle exe:bundle1"""

bundleEnd = '''
endBundle
endDocument'''


def createInputEntities(inputList, activity):
    '''
    Creates the input entity section of a PROV-N template specific to a python function
    :param inputList: List of names of the inputs to a function
    :param activity: Name of the function
    :param global_tracking: Whether global variables are being recorded
    :return: String defining the input entities and their relation to the function in PROV-N format
    '''
    activity = escape_forbidden_characters(activity)
    inputEntities = []
    inputList = [escape_forbidden_characters(x) for x in inputList]
    for i,inVal in enumerate(inputList):
        entityString = """
        entity(var:input{0}, [prov:value = 'var:input{0}value'])
        used(var:message, var:input{0}, -, [prov:role='exe:{2}_{1}'])""".format(i, inVal, activity)
        
        inputEntities.append(entityString)

    inputEntityString = '\n'.join(inputEntities)
    return inputEntityString


def createActivity(activity, agent):
    '''
    Creates the activity and agent string for the PROV-N code and the relations between them, corresponding to the
    function and python code, respectively
    :param activity: Name of function
    :param agent: Type of agent (typically software agent)
    :return: PROV-N format string for agent, activity and relations
    '''
    activity = escape_forbidden_characters(activity)
    agent = escape_forbidden_characters(agent)
    activityString = """
    activity(var:message, var:messageStartTime, var:messageEndTime, [prtr:activityName = 'var:activityName',
     prtr:hadMemoryUsage = 'var:memoryUsage', prtr:activitySource = 'var:moduleName', prov:label="var:comment" ])
    agent(var:lifeline, [prtr:modules = 'var:modules', prtr:python_version = 'var:python_version', prov:type='prtr:{1}'])

    wasAssociatedWith(var:message, var:lifeline, - , [])
    """.format(activity, agent)
    
    return activityString


def createActivityStarted(activity, agent):
    '''
    Creates the activity and agent string for the PROV-N code and the relations between them, corresponding to the
    function and python code, respectively. For the case when a function was started by another
    :param activity: Name of function
    :param agent: Type of agent (typically software agent)
    :return: PROV-N format string for agent, activity and relations
    '''

    activity = escape_forbidden_characters(activity)
    agent = escape_forbidden_characters(agent)
    activityString = """
    activity(var:message, var:messageStartTime, var:messageEndTime, [prtr:activityName = 'var:activityName',
     prtr:hadMemoryUsage = 'var:memoryUsage', prtr:activitySource = 'var:moduleName', prov:label="var:comment" ])
    activity(var:message2, var:message2StartTime, var:message2EndTime)
    agent(var:lifeline, [prtr:modules = 'var:modules', prtr:python_version = 'var:python_version', prov:type='prtr:{1}'])

    wasAssociatedWith(var:message, var:lifeline, - , [])
    wasStartedBy(var:message2, -, var:message, -)
    """.format(activity, agent)
    
    return activityString


def createOutputEntities(outputList):
    '''
    Creates the output entity section of a PROV-N template specific to a python function
    :param outputList: List of names of the outputs to a function
    :param global_tracking: Whether global variables are being recorded
    :return: String defining the output entities and their relation to the function in PROV-N format
    '''
    outputEntities = []
    outputList = [escape_forbidden_characters(x) for x in outputList]

    for o in range(len(outputList)):
        outputString = """
        entity(var:output{0}, [prov:value = 'var:output{0}value'])
        wasGeneratedBy(var:output{0}, var:message, -)""".format(o)

        outputEntities.append(outputString)

    outputEntityString = '\n'.join(outputEntities)

    return outputEntityString


def generateBundle(function, bundleStart, bundleEnd):
    '''
    Combines the PROV-N strings created by the other functions in this file and generates the full PROV-N template in
    string format.
    :param function: List generated by praetor which includes the function name, inputs, outputs, etc.
    :param bundleStart: PROV-N format string which declares the start of the provenance document
    :param bundleEnd: PROV-N format string which declares the end of the provenance document
    :return: String format PROV-N template for target function
    '''
    inEntities = createInputEntities(function[1], function[0])
    activity = createActivity(function[0],'SoftwareAgent')
    activityStarted = createActivityStarted(function[0],'SoftwareAgent')
    outEntities = createOutputEntities(function[2])

    bundleList = [bundleStart,inEntities,activity,outEntities,bundleEnd]
    bundle = '\n'.join(bundleList)
    
    bundleList2 = [bundleStart,inEntities,activityStarted,outEntities,bundleEnd]
    bundle2 = '\n'.join(bundleList2)

    return bundle, bundle2


def globalBundle(parameter_key, parameter_value, parameterID, parameter_type, activityID):
    '''
    Function used only when global tracking is enabled. Takes the names, values, and IDs of global parameters and
    generates PROV-N entity definitions for them as well as relations to the intended object already present in the
    provenance
    :param parameter_key: List of keys of the parameters in the global dictionary
    :param parameter_value: List of values of the parameters in the global dictionary
    :param globalID: Prov ID generated for the collection of global parameters
    :param parameterID: List of Prov IDs generated for the global dictionary
    :param parameter_type: List of python types for objects in the global dictionary
    :return: String representation of the entity definitions and relations for global variables
    '''
    global_bundle_relation = ''
    for key, val, pid, ptype in zip(parameter_key, parameter_value, parameterID, parameter_type):
        global_bundle_relation += '''
        entity({0}, [prov:value = "{2}" %% {3}, prov:type = "prtr:global_variable"])
        used({4}, {0}, -, [prov:role = "{1}"])
        '''.format(pid, key, val, ptype, activityID)

    return global_bundle_relation



#!/usr/bin/env python
# coding: utf-8

from collections import Counter
from datetime import datetime
import logging
import os
import requests

import matplotlib.pyplot as plt
import pandas as pd
try:
    from pandas.io.json import json_normalize
except:
    from pandas import json_normalize


DATABASE_HOST_URL = os.environ.get('DATABASE_HOST_URL', 'http://127.0.0.1:3030/')
REPOSITORY_ID = os.environ.get('REPOSITORY_ID', 'ds')
SPARQL_REST_URL = DATABASE_HOST_URL + REPOSITORY_ID

prefixes = '''
        PREFIX prov: <http://www.w3.org/ns/prov#>
        PREFIX run: <http://example.org/>
        PREFIX exe: <http://example.org/>
        PREFIX prtr: <https://praetor.pages.mpcdf.de/praetor_provenance/>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#> 
'''

prefix_dict = {'http://www.w3.org/ns/prov#': 'prov:',
               'http://example.org/': 'run:',
               'https://praetor.pages.mpcdf.de/praetor_provenance/': 'prtr:',
               'http://www.w3.org/2000/01/rdf-schema#': 'rdfs:'}

def remove_name_spaces(string, name_spaces):
    '''
    Remove namespaces from values in the provenance
    :param string: Input string to remove prefixes from
    :param name_spaces: Namespaces to remove
    :return: input string minus namespaces
    '''
    for x in name_spaces:
        try:
            string = string.replace(x, '')
        except AttributeError:
            for i, val in enumerate(string):
                string[i] = val.replace(x, '')
    return string


def clear_response(arr):
    logging.debug('StaticUtils: clearing given array')
    unwanted_strings = ['s', '', ' ', 's,p', 's,p,o', 'p,o', 'o', 'p', 'g', 'g,s,p,o', 's,p,o,r',
                        'activity,entity,p,dt,value,role,startTime,endTime', 's,s_count', 's,s_count,graphs', 's_count',
                        's,s_count,p,o,o_count', 'uuid,s,p,o', 's,g,s_count', 's,s_count,roles', 's,ts,te', 's,u']

    for string in unwanted_strings:
        if string in arr:
            arr.remove(string)

    logging.debug('StaticUtils: cleared list: ' + str(arr))

    return arr


def query_handler(query):
    '''
    Sends query to triple store
    :param query:
    :return: response of query
    '''
    response = requests.post(SPARQL_REST_URL,
                             data={'query': query},
                             timeout=86400)

    return response


def query_handler_histograms(query):
    '''
    Sends query to triple store
    :param query:
    :return: response of query
    '''
    response = requests.post(SPARQL_REST_URL,
                             data={'query': query},
                             headers={'Accept': 'text/csv'},
                             timeout=86400)

    return response

def upload_provenance(file_name):
    '''
    Function for uploading provenance files to the database
    :param file_name: name of provenance file to upload
    :return: Name of the graph in the database
    '''
    file_name_short = file_name.split('/')[-1]
    pipeline = 'http://' + file_name_short.replace('.ttl', '')
    data = open(file_name).read()
    headers = {'Content-Type': 'text/turtle;charset=utf-8'}
    url = SPARQL_REST_URL + '/data'
    requests.post(url, params={'graph': pipeline}, data=data, headers=headers)
    return pipeline


# def convert_to_datetime(datetime_str):
#     datetime_str = datetime_str[:-6]
#     time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%f')
#     return time


def convert_to_datetime_exception(datetime_str):
    '''
    Converts a string representation of datetime to a datetime object - temperamental
    :param datetime_str:
    :return: datetime object
    '''
    try:
        time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        try:
            time = datetime.strptime(datetime_str[:-6], '%Y-%m-%dT%H:%M:%S.%f')
        except ValueError:
            try:
                time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%fZ')
            except ValueError:
                try:
                    time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S')
                except:
                    time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%SZ')


    return time


def result_to_df(response, pre_dict=prefix_dict):
    '''
    Converts a response from the database to a pandas dataframe
    :param response: response to convert
    :return: pandas dataframe format response
    '''
    df = json_normalize(response.json()['results']['bindings'], meta='value')
    cols = [c for c in df.columns if c[-6:] == '.value']
    df2 = df[cols]
    col_names = [x for x in cols]

    if 'startTime.value' in col_names and 'endTime.value' in col_names:
        start_strs = df['startTime.value'].tolist()
        end_strs = df['endTime.value'].tolist()

        start_str = [x for x in start_strs]
        datetime_start = [convert_to_datetime_exception(x) for x in start_str]

        end_str = [x for x in end_strs]
        datetime_end = [convert_to_datetime_exception(x) for x in end_str]

        durations = [(x - y).total_seconds() for x, y in zip(datetime_end, datetime_start)]

        df2.insert(1, 'duration (s)', durations)

    df2 = df2.drop_duplicates()
    for key, value in pre_dict.items():
        df2 = df2.replace(key, value, regex=True)
    return df2


def user_defined_query(query):
    '''
    Function to pass a SPARQL query to the database
    :param query: string SPARQL query
    :return: dataframe of the results
    '''
    res = query_handler(query)
    response_as_string = clear_response(res.text.split("\r\n"))
    print(response_as_string)
    df = result_to_df(res)
    return df


def object_histogram(pipeline_list, object_name, object_type, prefixes=prefixes):
    '''
    Creates a histogram for the potential values for parameters or runtime metrics within one or more provenance datasets
    :param pipeline_list: list of provenance files to upload and query
    :param object_name: Name (role) of the parameter or name of the function to investigate
    :param object_type: can be 'entity' or 'activity'
    :param prefixes: prefixes required for the query
    :return: Data frame of the results and histogram plots
    '''
    res_string = ''
    frames = []
    for pipeline in pipeline_list:
        pipeline_name = upload_provenance(pipeline)

        if object_type == 'entity':
            query = prefixes + '''
            SELECT ?value
            FROM NAMED <''' + pipeline_name + '''>
            WHERE {
            GRAPH ?g {
            ?be a prov:Usage .
            ?be prov:hadRole ''' + object_name + ''' .
            ?be prov:entity ?entity .
            ?entity prov:value ?value .
            }
            }
            '''
        if object_type == 'activity':
            query = prefixes + '''
            SELECT ?startTime ?endTime ?memory ?output
            FROM NAMED  <''' + pipeline_name + '''>
            WHERE {
            GRAPH ?g {
            ?a a prov:Activity .
            ?a prtr:activityName  "''' + object_name + '''" . 
            ?a prov:startedAtTime ?startTime . 
            ?a prov:endedAtTime ?endTime .
            ?a prtr:hadMemoryUsage ?memory . 
            ?o prov:wasGeneratedBy ?a .
            ?o prov:value ?output .
            }
            }
            '''

        res = query_handler(query)
        df = result_to_df(res)
        df.to_csv('{}.csv'.format(pipeline))
        frames.append(df)
    final_df = pd.concat(frames)
    time_d = {}
    for col in final_df.columns:
        if col in ['startTime.value', 'endTime.value']:
            time_d[col] = final_df[col].values.tolist()

        else:
            if col != 'duration (s)':
                values = final_df[col].values.tolist()
                letter_counts = Counter(values)
                df = pd.DataFrame.from_dict(letter_counts, orient='index')
                ax = df.plot(kind='bar')
                ax.set_xlabel(col[:col.find('.')])
                ax.set_ylabel('Occurence')
                plt.show()

    if time_d != {}:
        duration = [(convert_to_datetime_exception(x) - convert_to_datetime_exception(y)).total_seconds() for x, y in
                    zip(time_d['endTime.value'], time_d['startTime.value'])]
        plt.hist(duration)
        plt.ylabel('Occurence')
        plt.xlabel('Duration')
        plt.xticks(rotation=90)
        plt.show()


    return final_df


def query_activity(pipeline, prefixes=prefixes):
    '''
    Queries the database for all activities within it and returns their names, ID, start/end times and memory usage
    :param pipeline: the name of the graph to query
    :param prefixes: prefixes required for the query
    :return: results from the query
    '''
    query = prefixes + '''

         SELECT ?activityID ?functionName ?startTime ?endTime ?memory ?functionSource 
         FROM NAMED <''' + pipeline + '''>                      

         WHERE {
         GRAPH ?g {
         
         ?activityID a prov:Activity .
         ?activityID prtr:activityName ?functionName.
         ?activityID prov:startedAtTime ?startTime.
         ?activityID prov:endedAtTime ?endTime.
         ?activityID prtr:hadMemoryUsage ?memory.
         ?activityID prtr:activitySource ?functionSource.
         }
         }

         '''

    res = query_handler(query)
    df = result_to_df(res)
    return df



def query_started_individual(pipeline, activity, prefixes=prefixes):
    '''
    A query to return all functions that were started by a target function
    :param pipeline: name of the graph to query
    :param activity: name of the activity
    :param prefixes: prefixes required for query
    :return: The identifier of each function started by the target activity
    '''

    if 'uuid:' in activity:
        activity = activity[activity.find('uuid:')+len('uuid:'):]

    query = prefixes + '''

         SELECT ?a
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?a2 prov:qualifiedStart ?b .
         ?b prov:hadActivity ?a .
        FILTER(CONTAINS(str(?a2), "''' + activity + '''"))
         }
         }
         '''
    res = query_handler(query)
    response_as_string = clear_response(res.text.split("\r\n"))
    df = result_to_df(res)
    return df


def query_was_started_by_individual(pipeline, activity, prefixes=prefixes):
    '''
    A query to return all functions that started the target function
    :param pipeline: name of the graph to query
    :param activity: name of the activity
    :param prefixes: prefixes required for query
    :return: The identifier of each function that started the target activity
    '''
    query = prefixes + '''

         SELECT ?a
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?a prov:qualifiedStart ?b .
         ?b prov:hadActivity ?a2 .
         FILTER(CONTAINS(str(?a2), "''' + activity + '''"))

         }
         }
         '''
    res = query_handler(query)
    df = result_to_df(res)
    return df


def query_started(pipeline, prefixes=prefixes):
    '''
    A query to return all functions that were started by another
    :param pipeline: name of the graph to query
    :param prefixes: prefixes required for query
    :return: The identifier of each function started by another
    '''
    query = prefixes + '''

         SELECT ?startedBy ?started
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {
        
         ?startedBy prov:qualifiedStart ?awsb .
         ?awsb prov:hadActivity ?started.

         }
         }
         '''
    res = query_handler(query)
    df = result_to_df(res)
    return df


def query_input(pipeline, prefixes=prefixes):
    '''
    A query to find all inputs within a provenance dataset, the results show their ID, role, and value
    :param pipeline: the name of the graph to query
    :param prefixes: prefixes required for query
    :return: The identifier, role, and value of every entity used as an input
    '''
    query = prefixes + '''
         
         SELECT ?activityID ?entityID ?objectName ?value
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?activityID prov:qualifiedUsage ?ni .
         ?ni prov:entity ?entityID .
         ?ni prov:hadRole ?objectName .
         ?entityID prov:value ?value .

         }
         }
         '''
    res = query_handler(query)
    df = result_to_df(res)
    return df


def query_output(pipeline, prefixes=prefixes):
    '''
    A query to find all outputs within a provenance dataset, the results show their ID, role, and value
    :param pipeline: the name of the graph to query
    :param prefixes: prefixes required for the query
    :return: The identifier, role, and value of every entity used as an output
    '''
    query = prefixes + '''
         SELECT ?activityID ?value ?objectName 
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?objectName prov:wasGeneratedBy ?activityID .
         ?objectName prov:value ?value. 

         }
         }
         '''
    res = query_handler(query)
    df = result_to_df(res)
    return df


def query_output_individual(pipeline, activity, prefixes=prefixes):
    '''
     A query to find all outputs for a target function within a provenance dataset, the results show their ID, role,
      and value
    :param pipeline: name of the graph to query
    :param activity: name of the activity
    :param prefixes: prefixes required for query
    :return: The identifier, role, and value of every entity used as an output in target function
    '''
    query = prefixes + '''

         SELECT ?activityID ?entityID ?value
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?entityID prov:wasGeneratedBy ?activityID .
         ?activityID prtr:activityName ''' + activity + ''' .
         ?entityID prov:value ?value . 
         

         }
         }
         '''

    res = query_handler(query)
    df = result_to_df(res)
    return df


def query_input_individual(pipeline, entity, prefixes=prefixes):
    '''
    A query to find all of a specific input within a provenance dataset, the results show their ID, role, and value
    :param pipeline: the name of the graph to query
    :param entity: role of an entity to search (name of the entity in python)
    :param prefixes: prefixes required for the query
    :return: The identifier, role, and value of every entity used as an input
    '''
    query = prefixes + '''

         SELECT ?activityID ?entityID ?value
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?activityID prov:qualifiedUsage ?ni .
         ?ni prov:entity ?entityID .
         ?ni prov:hadRole ''' + entity + ''' .
         ?entityID prov:value ?value .

         }
         }
         '''

    res = query_handler(query)
    df = result_to_df(res)
    return df

def query_quality(pipeline, prefixes=prefixes):
    '''
    A query to extract quality metrics from provenance datasets
    :param pipeline: the name of the graph to query
    :param prefixes: prefixes required for the query
    :return: All values and metrics of quality contained within the provenance and what they are attached to
    '''

    query = prefixes + '''
    SELECT ?qualityMetric ?qualityValue ?qualityID 
    FROM NAMED <''' + pipeline + '''>
    
    WHERE {
    GRAPH ?g {

    ?qid a prov:Entity ;
        rdfs:label ?qualityMetric ;
        rdfs:subClassOf "prtr:QualityMeasure" .
        
    ?qualityID a prov:Entity ;
        prov:value ?qualityValue ;
        a ?qq .
    
    FILTER(CONTAINS(STR(?qq), ":quality_measure_"))
    
    }
    }
    '''

    query_2 = prefixes + '''
    SELECT ?qualityID ?objectID
    FROM NAMED <''' + pipeline + '''>
    
    WHERE {
    GRAPH ?g {

    ?objectID prtr:hadQuality ?qualityID .
    
    }
    }
    
    '''

    res = query_handler(query)
    df = result_to_df(res)
    res = query_handler(query_2)
    df2 = result_to_df(res)
    df_comb = df.merge(df2, how='outer', on='qualityID.value')
    df_comb = df_comb.dropna()
    return df_comb


def query_resources_used(pipeline, activity=None, prefixes=prefixes):
    '''
    Calculates the total memory and time for a function or a pipeline as a whole
    :param pipeline: name of the graph to query
    :param activity: name of the activity
    :param prefixes: prefixes required for query
    :return: total time (s) and memory (MB)
    '''
    query = prefixes + '''
    
    
         SELECT ?activityID ?startTime ?endTime ?memory 
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {
         ?activityID a prov:Activity .
         ?activityID prov:startedAtTime ?startTime.
         ?activityID prov:endedAtTime ?endTime.
         ?activityID prtr:hadMemoryUsage ?memory
         }
         }
    
    '''
    res = query_handler(query)
    results = result_to_df(res)

    if activity != None:
        sorted_df = results.sort_values(by=['startTime'], ascending=True)
        sorted_df = sorted_df.reset_index(drop=True)
        start = sorted_df.loc[sorted_df['activityID'] == activity].index.tolist()
        results = results[:start[0]]

    func_times = results['duration (s)'].tolist()

    memory = results['memory.value'].tolist()
    mem_vals = [float(x) for x in memory]

    tot_time = sum(func_times)
    func_mem = [(x * y) for x, y in zip(func_times, mem_vals)]
    tot_mem = sum(func_mem)

    return {'total_time_s': tot_time, 'total_memory_MB': tot_mem}


def query_av_mem(pipeline, prefixes=prefixes):
    '''
    Determine all function which had a greater than average memory consumption
    :param pipeline: Name of pipeline to query
    :param prefixes: Prefixes  required for query
    :return: Names of all functions with greater than average memory usage
    '''
    query = prefixes + '''

         SELECT ?n 
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?a a prov:Activity .
         ?a prtr:activityName ?n . 
         ?a prtr:hadMemoryUsage ?mem.
         FILTER( ?mem > AVG(?mem))
         
         }
         }
         GROUP BY ?a
         '''

    res = query_handler(query)
    return res



def query_activity_source(pipeline, module, prefixes=prefixes):
    '''
    Query to find all activities belonging to a particular source, typically either main or a module name
    :param pipeline: pipeline name to query
    :param module: name of module
    :param prefixes: prefixes required for query
    :return: All activities that belong to the target module
    '''
    query = prefixes + '''

         SELECT ?a ?an
         FROM NAMED <''' + pipeline + '''>

         WHERE {
         GRAPH ?g {

         ?a a prov:Activity .
         ?a prtr:activityName ?an .
         ?a prtr:activitySource ''' + module + ''' .
         }
         }

         '''

    res = query_handler(query)

    return res


def forward_search(pipeline, object_id, n_itter, prefixes=prefixes):
    '''
    Function to find the n next objects in the provenance that are connected to object_id
    :param pipeline: name of provenance graph to query over
    :param object_id: id of object for the start of the search
    :param n_itter: number of connections from initial object to explore
    :param prefixes: prefixes required for the query
    :return: dataframe of all objects connected to the object_id over n connections
    '''
    activity_id = False
    structure_1 = '''
    FROM NAMED <''' + pipeline + '''>
    
    WHERE {
    GRAPH ?g {
    
    '''

    structure_2 = '''
    }
    }
    '''
    if 'uuid:' in object_id:
        object_id_search = object_id[object_id.find('uuid:')+len('uuid:'):]
        activity_id = True
    selects = ['?node{0}'.format(x) for x in range(n_itter+1)]
    sel_str = 'SELECT ' + ' '.join(selects)
    if activity_id:
        main_con = '?a ?relation0 ?node0 .'
    else:
        main_con = '{} ?relation0 ?node0 .'.format(object_id)
    connections = ['?node{0} ?relation{1} ?node{1} .'.format(x, x+1) for x in range(n_itter)]
    con_str = '\n'.join(connections)
    query =  main_con + '\n' + con_str
    filter_start = 'FILTER('
    filter_str = ['?relation{} != prov:wasAssociatedWith'.format(x) for x in range(n_itter)]
    filter_string = ' && '.join(filter_str)
    if activity_id:
        query += '\n' + filter_start + filter_string + ' && CONTAINS(STR(?a), "{}")'.format(object_id_search) + ') .\n'
    else:
        query += '\n' + filter_start + filter_string +') .\n'
    final_query = prefixes + sel_str + structure_1 + query + structure_2
    res = query_handler(final_query)
    df = result_to_df(res)
    df = df.drop_duplicates()
    return df


def backward_search(pipeline, object_id, n_itter, prefixes=prefixes):
    '''
    Function to find the n previous objects in the provenance that are connected to object_id
    :param pipeline: name of provenance graph to query over
    :param object_id: id of object for the start of the search
    :param n_itter: number of connections from initial object to explore
    :param prefixes: prefixes required for the query
    :return: dataframe of all objects connected to the object_id over n connections
    '''
    activity_id = False
    structure_1 = '''
    FROM NAMED <''' + pipeline + '''>

    WHERE {
    GRAPH ?g {

    '''

    structure_2 = '''
    }
    }
    '''
    if 'uuid:' in object_id:
        object_id_search = object_id[object_id.find('uuid:') + len('uuid:'):]
        activity_id = True
    selects = ['?node{0}'.format(x) for x in range(n_itter)]
    sel_str = 'SELECT ' + ' '.join(selects)
    if activity_id:
        main_con = '?node0 ?relation0 ?a'
    else:
        main_con = '?node0 ?relation0 {} .'.format(object_id)
    connections = ['?node{1} ?relation{1} ?node{0} .'.format(x, x + 1) for x in list(reversed(range(n_itter)))[1:]]
    con_str = '\n'.join(connections)
    query = con_str + '\n' + main_con
    filter_start = 'FILTER('
    filter_str = ['?relation{} != prov:wasAssociatedWith'.format(x) for x in range(n_itter)]
    filter_string = ' && '.join(filter_str)
    if activity_id:
        '\n' + filter_start + filter_string + ' && CONTAINS(STR(?a), "{}")'.format(object_id_search) + ') .\n'
    else:
        query += '\n' + filter_start + filter_string + ') .\n'

    final_query = prefixes + sel_str + structure_1 + query + structure_2
    res = query_handler(final_query)
    df = result_to_df(res)
    df = df.drop_duplicates()
    return df



#!/usr/bin/env python
# coding: utf-8

from collections import Counter
import datetime
import functools
import hashlib
import importlib
import inspect
import json
import os
import subprocess
import sys
import textwrap
import time
import types
import uuid

try:
    from _thread import get_ident
except ImportError:
    from _dummy_thread import get_ident

try:
    prtr_path = os.environ['PRAETOR']
    sys.path.append(prtr_path)
    import praetor_settings_user as praetor_settings
except:
    from praetor import praetor_settings

sys.path.append(praetor_settings.provenance_directory+'{}/'.format(os.environ['PRAETOR_pipeline_id']))

from praetor import praetorTemplates

import templateDictionary


if praetor_settings.memory:
    from memory_profiler import memory_usage

praetor_settings.blacklist.extend(['haslib.md5', 'datetime.datetime.now','inspect.signature', 'inspect.Parameter.empty',
                                   'final_globals.items','uuid.uuid4',
                                   'json.dump', 'inspect.ismodule', 'inspect.isfunction', 'inspect.isclass',
                                   'inspect.unwrap', 'inspect.getsource', 'importlib.reload', 'inspect.signature',
                                   'inspect.stack', 'urandom', 'os.fspath', 'os.path.normcase', 'os.getcwd',
                                   'os.fsencode', 'callable', 'inspect.from_callable', 'inspect.signature', 'getattr',
                                   'isinstance', 'print', 'os.', 'abc.__instancecheck__', '_collections_abc.__eq__',
                                   '_abc._abc_instancecheck_', '_collections_abc.__eq__', 'posix.pipe', 'posix.read',
                                   'posix.close', 'collections.abc._check_methods', '_abc._abc_instancecheck',
                                   '_abc._abc_subclasscheck', 'os.get_exec_path', 'posix.lstat',
                                   'collections.namedtuple', 'undefined.__contains__', 'functools.wraps',
                                   'functools.update_wrapper', 'posix.register_at_fork', 'os._get_exports_list',
                                   '_abc._abc_init', '_abc._abc_init', 'abc.abstractmethod', 'posix.getpid',
                                   'posix.confstr', 'posix.readlink', 'undefined', 'functools.lru_cache', '__contains__',
                                   'collections._check_methods_collections.abc', '_check_methods'])


ts = time.time()

bindings = {
    "context": {"xsd": "http://www.w3.org/2001/XMLSchema#", "lsim": "http://example.org/", "urn_uuid": "urn:uuid:",
                "run": "http://example.org/"}}

session_id = 'praetor'
prov_id_counter = Counter()
prov_id_cache = dict()
prov_id_activity_cache = []
call_func_list = []
glob_vars = [{}]
praetor_rec = 0

dtypes = {'int': 'int', 'unsignedInt': 'unsignedInt', 'hexBinary': 'hexBinary', 'NOTATION': 'NOTATION',
          'nonPositiveInteger': 'nonPositiveInteger', 'float': 'float', 'ENTITY': 'ENTITY', 'bool': 'boolean',
          'positiveInteger': 'positiveInteger', 'duration': 'duration', 'IDREFS': 'IDREFS',
          'unsignedLong': 'unsignedLong', 'normalizedString': 'normalizedString', 'dateTimeStamp': 'dateTimeStamp',
          'NMTOKEN': 'NMTOKEN', 'negativeInteger': 'negativeInteger', 'base64Binary': 'base64Binary',
          'long': 'long', 'unsignedShort': 'unsignedShort', 'ENTITIES': 'ENTITIES', 'anyURI': 'anyURI',
          'NMTOKENS': 'NMTOKENS', 'IDREF': 'IDREF', 'unsignedByte': 'unsignedByte', 'Name': 'Name',
          'dayTimeDuration': 'dayTimeDuration', 'date': 'date', 'integer': 'integer', 'byte': 'byte', 'ID': 'ID',
          'gMonth': 'gMonth', 'short': 'short', 'language': 'language', 'gMonthDay': 'gMonthDay',
          'double': 'double', 'Decimal': 'decimal', 'gDay': 'gDay', 'gYearMonth': 'gYearMonth',
          'QName': 'QName', 'datetime': 'dateTime', 'nonNegativeInteger': 'nonNegativeInteger', 'gYear': 'gYear',
          'token': 'token', 'time': 'time', 'yearMonthDuration': 'yearMonthDuration', 'NCName': 'NCName',
          'str': 'string'}



def md5(fname):
    '''
    Function to create hash codes that represent data, files, etc.
    :param fname: Name of file to read and hash
    :return: Hash code specific to file fname
    '''
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()


def get_agent_string(modules):
    '''
    Create unique id for the agent of a pipeline and get the version of python and all imported modules
    :param modules: List of module names imported
    :return: Unique id for the agent and list of versions for modules
    '''
    versions = []
    try:
        procPy = subprocess.Popen(['python3', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except:
        procPy = subprocess.Popen(['python', '--version'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    pythonVersion = procPy.stdout.read().decode('utf-8')
    loc = 'Python '
    pyVer = pythonVersion[pythonVersion.find(loc)+len(loc):]
    pyVer = pyVer.strip()
    pyVer = ''.join(e for e in pyVer if e.isalnum() or e == '.' or e == '-')
    pyVer = 'run:'+pyVer
    
    for mod in modules:
        try:
            proc1 = subprocess.Popen(['pip3', 'show',mod], stdout=subprocess.PIPE)
        except:
            proc1 = subprocess.Popen(['pip', 'show', mod], stdout=subprocess.PIPE)
        proc2 = subprocess.Popen(['grep', 'Version'], stdin=proc1.stdout, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        proc1.stdout.close()

        out, err = proc2.communicate()
        versions.append(out.decode('utf-8'))

    versions = [x.replace('\n', '') for x in versions]
    versions = [x[x.find(': ')+2:] for x in versions]

    modString = [x + '_' + y for x, y in zip(modules, versions)]
    modString = ', '.join(modString)
    return modString, pyVer


def get_modules():
    '''
    Function to create the agent_json.json file, including creating a unique id for the agent, determining all modules
    imported, find their versions, structure and input all infroamtion into agent_json.json
    :return: agent_json.json
    '''
    modules = []
    for name, val in globals().items():
        if isinstance(val, types.ModuleType):
            modName = val.__name__
            loc = modName.find('.')
            if loc != -1:
                modName = val.__name__[0:loc]
            modules.append(modName)

    exclude = ['__builtin__',
               'Tester','_thread','graphlib','scriptScript','__future__','_threading_local','grp','scriptresubmitter',
               '_abc','_tkinter','gzip','secrets','_aix_support','_tracemalloc','hashlib','select','_ast','_uuid',
               'heapq','selectors','_asyncio','_warnings','hmac','setuptools','_bisect','_weakref','html','shelve',
               '_blake2','_weakrefset','http','shlex','_bootlocale','_xxsubinterpreters','idlelib','shutil',
               '_bootsubprocess','_xxtestfuzz','imaplib','signal','_bz2','_zoneinfo','imghdr','site','_codecs','abc',
               'imp','smtpd','_codecs_cn','aifc','importlib','smtplib','_codecs_hk','antigravity','inspect','sndhdr',
               '_codecs_iso2022','argparse','io','socket','_codecs_jp','array','ipaddress','socketserver','_codecs_kr',
               'ast','itertools','spwd','_codecs_tw','asynchat','json','sqlite3','_collections','asyncio','keyword',
               'sre_compile','_collections_abc','asyncore','latexWriter','sre_constants','_compat_pickle','atexit',
               'lib2to3','sre_parse','_compression','audioop','linecache','ssl','_contextvars','base64','locale',
               'stat','_crypt','bdb','logging','statistics','_csv','binascii','lzma','string','_ctypes','binhex',
               'mailbox','stringprep','_ctypes_test','bisect','mailcap','struct','_curses','builtins','marshal',
               'subprocess','_curses_panel','bwd','math','sunau','_datetime','bz2','mimetypes','symbol','_decimal',
               'cProfile','mmap','symtable','_distutils_hack','calendar','modulefinder','sys','_elementtree','cgi',
               'multiprocessing','sysconfig','_functools','cgitb','netrc','syslog','_hashlib','chunk','nis','tabnanny',
               '_heapq','cmath','nntplib','tarfile','_imp','cmd','ntpath','telnetlib','_io','code','nturl2path',
               'tempfile','_json','codecs','numbers','termios','_locale','codeop','opcode','test','_lsprof',
               'collections','operator','tester','_lzma','colorsys','optparse','testing_prtr_packagetemplateDictionary',
               '_markupbase','compileall','os','textwrap','_md5','concurrent','ossaudiodev','this','_multibytecodec',
               'configparser','parser','threading','_multiprocessing','contextlib','pathlib','time','_opcode',
               'contextvars','pdb','timeit','_operator','copy','periodic','tkinter','_osx_support','copyreg','pickle',
               'token','_peg_parser','crypt','pickletools','tokenize','_pickle','csv','pip','trace','_posixshmem',
               'ctypes','pipes','traceback','_posixsubprocess','curses','pkg_resources','tracemalloc','_py_abc',
               'dataclasses','pkgutil','tty','_pydecimal','datetime','platform','turtle','_pyio','dbm','plistlib',
               'turtledemo','_queue','decimal','poplib','types','_random','difflib','posix','typing','_sha1','dis',
               'posixpath','unicodedata','_sha256','distutils','pprint','unittest','_sha3','doctest','profile','urllib',
               '_sha512','email','provcall-NoWorkflow-save','uu','_signal','encodings','pstats','uuid','_sitebuiltins',
               'ensurepip','pty','valuefilemaker','_socket','enum','pwd','venv','_sqlite3','errno','py_compile',
               'warnings','_sre','faulthandler','pyclbr','wave','_ssl','fcntl','pydoc','weakref','_stat','filecmp',
               'pydoc_data','webbrowser','_statistics','fileinput','pyexpat','wheel','_string','fnmatch','pytest',
               'wsgiref','_strptime','formatter','python_modules','xdrlib','_struct','fractions','queue','xml',
               '_symtable','ftplib','quopri','xmlrpc','_sysconfigdata__linux_x86_64-linux-gnu','functools',
               'random','xxlimited','_sysconfigdata_x86_64_conda_cos6_linux_gnu','gc','re','xxsubtype',
               '_sysconfigdata_x86_64_conda_linux_gnu','genBindings','readline','zerosigmaker','_testbuffer',
               'genericpath','reprlib','zipapp','_testcapi','getopt','resource','zipfile','_testimportmultiple',
               'getpass','rlcompleter','zipimport','_testinternalcapi','gettext','runpy','zlib','_testmultiphase',
               'glob','sched','zoneinfo',
               'genBindings','praetor','convert2prov','decorate','praetorTemplates','praetor_settings',
               'praetor_settings_user','templateDictionary', 'python_modules', 'uuid4']

    modules = [x for x in modules if x not in exclude]

    json_total = {}
    modules_versions, py_version = get_agent_string(modules)

    bindings['var'] = {}
    bindings['var']['modules'] = ['run: ({})'. format(modules_versions)]
    bindings['var']['python_version'] = [py_version]
    bindings['var']['lifeline'] = ['urn_uuid:{}'.format(os.environ["PRAETOR_pipeline_id"])]

    json_dir = praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/json/'

    json_total['agent'] = bindings

    with open(json_dir + 'agent_json.json', 'w') as f:
        json.dump(json_total, f)


def timestamp():
    ts = str(time.time())
    return ts


def datetimestamp():
    dts = datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')
    return dts


def gen_identifier(variable, naming_template="entity"):
    '''
    General utility function to create prov ids for python objects
    :param variable: Value of object to create id for
    :param naming_template: Name related to the object in question to include in the id
    :return: id for target object
    '''
    try:
        prov_id = prov_id_cache[id(variable)]
    except KeyError:
        prov_id_counter[naming_template] += 1
        prov_id = '{}_{}_{}'.format(naming_template, session_id, prov_id_counter[naming_template])
        prov_id_cache[id(variable)] = prov_id

    return prov_id


def escape_forbidden_characters(in_string):
    '''
    Escape forbidden characters in input string in accordance with PROV-N
    :param in_string: input string to escape characters within
    :return: modified in_string with escaped characters
    '''
    forbidden = '\\(=)|,:;[]'
    allowed_special = '/._-'
    no_forbidden = [x if x.isalnum() else '\\' + x if x in forbidden else x if x in allowed_special else '' for x in
                 in_string]
    final_string = ''.join(no_forbidden)
    return final_string


def remove_quotes_from_string(in_string):
    out_string = [x for x in in_string if x not in ['"', "'"]]
    out_string = ''.join(out_string)
    return out_string


def find_type(value):
    '''
    If possible, return the type of input value and convert it to string
    :param value: Input value to type
    :param praetor_rec: switch to turn off provenance recording
    :return: String value, type of value
    '''

    global praetor_rec
    try:
        praetor_rec += 1
        outVal = str(value)
        praetor_rec = 0
        outVal = remove_quotes_from_string(outVal)
        otype = type(value).__name__
        if otype in dtypes.keys():
            ptype = dtypes[otype]
        else:
            ptype = 'string'
    except:
        praetor_rec += 1
        outVal = str(value)
        praetor_rec = 0
        outVal = remove_quotes_from_string(outVal)
        ptype = 'string'
    return outVal, ptype


def reduce_lists(in_list):
    '''
    Sub-rountine for un-nesting lists if they are present
    :param in_list: Nested input list
    :return: Flattened version of in_list
    '''
    while any(isinstance(i, list) for i in in_list):
        flat_new = []
        for sublist in in_list:
            if not isinstance(sublist, list):
                flat_new.append(sublist)
            else:
                for value in sublist:
                    flat_new.append(value)
        in_list = flat_new

    return in_list


def get_default_args(func):
    '''
    Extract the source code for a wrapped function
    :param decorated: Function source code including decorator
    :return: Function source code excluding decorator
    '''
    try:
        signature = inspect.signature(func)
    except (TypeError, AttributeError, ValueError) as e:
        return [], []
    out_dict =  {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }
    keys = [x for x in out_dict.keys()]
    values = [x for x in out_dict.values()]
    return keys, values


def extract_wrapped(decorated):
    closure = (c.cell_contents for c in decorated.__closure__)
    return next((c for c in closure if isinstance(c, types.FunctionType)), None)


def extract_string_type(type_string):
    locater = "'"
    loc1 = type_string.find(locater)
    loc2 = type_string[loc1+len(locater):].find(locater)
    final_type = type_string[loc1+len(locater):loc1+len(locater)+loc2]
    return final_type


def get_class_that_defined_method(meth):
    try:
        for cls in inspect.getmro(meth.im_class):
            if meth.__name__ in cls.__dict__:
                return cls.__name__

    except AttributeError:
        return None
    return None


def find_dict_differences(dict1, dict2):
    '''
    Used to find the differences between two dictionaries of global variables
    :param dict1: initial global variable dictionary
    :param dict2: updated global variable dictionary
    :return: Items in dict2 that are not in dict1
    '''
    differences = {}

    for key in dict2.keys():
        if key not in dict1:
            differences[key] = (dict2[key])

    for key in set(dict1.keys()).intersection(dict2.keys()):
        try:
            if dict1[key] != dict2[key]:
                differences[key] = ( dict2[key])
        except ValueError:
            if str(dict1[key]) != str(dict2[key]):
                differences[key] = (dict2[key])

    return differences


def track_globs(initial_globals):
    '''
    Find changes in global variabes between and old set and a freshly generated one
    :param initial_globals: Set of global variables to compare to
    :return: Differences between current and previous global variables
    '''
    final_globals = dict(globals())
    changed_globals = {name: value for name, value in final_globals.items() if name not in initial_globals or initial_globals[name] != value}
    return changed_globals


def provcall(inlist, outlist, template, longname, startTime, endTime, memUse, json_name,
             json_out='full_json.json', jsonDir='./', maxValueLength=praetor_settings.max_value_length, module_name='main',
             global_inputs=None, global_outputs=None):
    '''
    Consumes as input the various parameters and metrics surrounding a function execution and writes them into a json
    file as bindings, ready to combine with a template and make provenance.
    :param inlist: List of input values
    :param outlist: List of output values
    :param template: Name of provenance template corresponding to the function
    :param longname: Full name of the function (String)
    :param startTime: Start time of execution (dateTime object)
    :param endTime: End time of execution (dateTime object)
    :param memUse: Average memory consumption during function (Bytes)
    :param json_name: String for key to identify function execution in json file
    :param json_out: Name of json file to write data to
    :param jsonDir: Directory where json_out is located
    :param maxValueLength: Max number of characters for an input/output before it is saved as a "big entity"
    :param module_name: Name of thing where the function is defined (either main or a module)
    :param global_inputs: Dictionary of global variables consumed by a function
    :param global_outputs: Dictionary of global variables produced by a function
    :return: Updates json_out with the above information for the target function execution
    '''

    global praetor_rec

    template = ''.join(e if e.isalnum() else e if e == '_' else '' for e in template)
    longname = ''.join(e if e.isalnum() else e if e == '_' else '' for e in longname)
    identifier = uuid.uuid4()
    json_total = {}
    bindings['var'] = {}
    bindings['var']['messageStartTime'] = [{"@type": "xsd:dateTime", "@value": startTime}]
    bindings['var']['moduleName'] = [{"@type": "xsd:string", "@value": module_name}]
    bindings['var']['activityName'] = [{"@type": "xsd:string", "@value": longname}]

    if len(call_func_list) != 0 and call_func_list[-1] != '<module>':
        bindings['var']['message2'] = [{"@id": prov_id_activity_cache[-1][0]}]
        bindings['var']['message2StartTime'] = [{"@type": "xsd:dateTime", "@value": prov_id_activity_cache[-1][1]}]
        bindings['var']['message2EndTime'] = [{"@type": "xsd:dateTime", "@value": prov_id_activity_cache[-1][2]}]
        bindings['var']['template'] = [template + 'started']
    else:
        bindings['var']['template'] = [template]
    bindings['var']['message'] = [{"@id": "urn_uuid:{}_{}".format(longname, identifier)}]
    prov_id_activity_cache.append(["urn_uuid:{}_{}".format(longname, identifier),startTime,endTime])

    if longname in templateDictionary.tempDict:
        try:
            if len(outlist) < len(templateDictionary.tempDict[longname][1]):
                lenDiff = len(templateDictionary.tempDict[longname][1]) - len(outlist)
                for extra in range(lenDiff):
                    outlist.append('Not Computed')
            if len(inlist) < len(templateDictionary.tempDict[longname][0]):
                lenDiff = len(templateDictionary.tempDict[longname][0]) - len(inlist)
                for extra in range(lenDiff):
                    inlist.append('Not Computed')
        except KeyError:
            pass

    longfileDir = jsonDir.replace('json', 'big_entities')

    for x in range(len(outlist)):
        if callable(outlist[x]):
            outVal, ptype = find_type('callable_{}'.format(type(outlist[x])))
            object_id = gen_identifier(outVal, naming_template='{}_outEntity_{}'.format(longname, x))
        else:
            outVal, ptype = find_type(outlist[x])
            object_id = gen_identifier(outlist[x], naming_template='{}_outEntity_{}'.format(longname, x))

        if len(outVal) > maxValueLength:
            with open(longfileDir + object_id + '.txt', 'w') as f:
                praetor_rec += 1
                f.write(str(outlist[x]))
                praetor_rec = 0
            outVal = longfileDir + object_id + '.txt'
            ptype = 'string'
        bindings['var']['output' + str(x)] = [{"@id": "run:" + object_id}]
        bindings['var']['output' + str(x) + 'value'] = [{"@value": outVal, "@type": "xsd:{}".format(ptype)}]

    if global_outputs is not None:
        object_id = gen_identifier(global_outputs, naming_template='globals')
        bindings['var']['praetorglobaloutput'] = [{"@id": "run:" + object_id}]
        for glob_key, glob_val in global_outputs.items():
            object_id = gen_identifier(glob_val, naming_template='globals_{}'.format(glob_key))
            g_val, ptype = find_type(glob_val)
            if len(g_val) > maxValueLength:
                with open(longfileDir + object_id + '.txt', 'w') as f:
                    f.write(str(glob_val))
                g_val = longfileDir + object_id + '.txt'
                ptype = 'string'
            bindings['var']['praetorglobaloutput{}'.format(glob_key)] = [{"@id": "run:" + object_id,
                                                                          "@value": g_val,
                                                                          "@type": "xsd:{}".format(ptype),
                                                                          "@label": 'run:{}'.format(glob_key)}]

    for x in range(len(inlist)):
        if callable(inlist[x]):
            inVal, ptype = find_type('callable_{}'.format(type(inlist[x])))
            object_id = gen_identifier(inVal, naming_template='{}_inEntity_{}'.format(longname, x))
        else:
            inVal, ptype = find_type(inlist[x])
            object_id = gen_identifier(inlist[x], naming_template='{}_inEntity_{}'.format(longname, x))

        if len(inVal) > maxValueLength:
            with open(longfileDir + object_id + '.txt', 'w') as f:
                praetor_rec += 1
                f.write(str(inlist[x]))
                praetor_rec = 0
            inVal = longfileDir + object_id + '.txt'
            ptype = 'string'

        bindings['var']['input' + str(x)] = [{"@id": "run:" + object_id}]
        bindings['var']['input' + str(x) + 'value'] = [{"@value": inVal, "@type": "xsd:{}".format(ptype)}]

    if global_inputs is not None:
        object_id = gen_identifier(global_inputs, naming_template='globals')
        bindings['var']['praetorglobalinput'] = [{"@id": "run:" + object_id}]
        for glob_key, glob_val in global_inputs.items():
            object_id = gen_identifier(glob_val, naming_template='globals_{}'.format(glob_key))
            g_val, ptype = find_type(glob_val)
            if len(g_val) > maxValueLength:
                with open(longfileDir + object_id + '.txt', 'w') as f:
                    f.write(str(glob_val))
                g_val = longfileDir + object_id + '.txt'
                ptype = 'string'
            bindings['var']['praetorglobalinput{}'.format(glob_key)] = [{"@id": "run:" + object_id, "@value": g_val,
                                                                         "@type": "xsd:{}".format(ptype),
                                                                         "@label": 'run:{}'.format(glob_key)}]

    bindings['var']['messageEndTime'] = [{"@type": "xsd:dateTime", "@value": endTime}]
    bindings['var']['memoryUsage'] = [{"@type": "xsd:string", "@value": memUse}]

    if len(call_func_list) != 0 and call_func_list[-1] != '<module>':
        json_total[json_name + 'started'] = bindings
        with open(jsonDir + json_out, 'a') as f:
            json.dump(json_total, f)
            f.write('\n')

    else:
        json_total[json_name] = bindings
        with open(jsonDir + json_out, 'a') as f:
            json.dump(json_total, f)
            f.write('\n')


def printDict():
    print(prov_id_cache, "prov_id_cache")


def get_globals():
    '''
    Creates a dictionary of all global variables present within the target pipeline, execluding those added by the
    praetor  code and built-in variables
    :return: Dictionary of global variables
    '''
    frame_1 = inspect.currentframe().f_back
    frame = frame_1.f_back
    globals_dict = frame.f_globals
    main_script_globals = {key:value for key,value in globals_dict.items() if not key.startswith('__')
                           and not any(sub in key for sub in ['genBindings', 'praetor', 'prov_id_counter'])
                           and not (inspect.ismodule(value) or inspect.isfunction(value) or inspect.isclass(value))}
    return main_script_globals


def create_template(func, template_name, template_name_started, function_name, filePath):
    '''
    Create a PROV-N template for a target python function
    :param func: Target function object
    :param template_name: Name of PROV-N template to create
    :param template_name_started: Name of PROV-N template if the function was started by another
    :param function_name: Name of function
    :param filePath: Path where templates will be generated
    :param global_tracking: Whether to track global variables
    :return: Creates the appropriate template for the function and stores a copy of the source code
    '''
    if isinstance(func, types.FunctionType):
        try:
            func_unwrapped = extract_wrapped(func)
        except (TypeError, AttributeError) as e:
            func_unwrapped = func
        func_source = inspect.unwrap(func_unwrapped)
        source = textwrap.dedent(inspect.getsource(func_source))
        source_str = source.encode('ascii', 'ignore').decode("utf-8")
        output_list = praetorTemplates.separate_output(source_str)
        try:
            input_list = [key for key in inspect.signature(func_source).parameters.keys()]
        except (TypeError, AttributeError, ValueError) as e:
            input_list = ['inputs_not_found']
        funct_name = func_source.__name__
        function_parameters = [funct_name, input_list, output_list]
        temp_sep_dict = {function_parameters[0]: [function_parameters[1], function_parameters[2]]}
        importlib.reload(templateDictionary)
        full_temp_dict = templateDictionary.tempDict
        full_temp_dict.update(temp_sep_dict)
        bundle, bundleStarted = praetorTemplates.generateBundle(function_parameters, praetorTemplates.bundleStart,
                                                                praetorTemplates.bundleEnd)

        with open(filePath + 'templates/' + template_name, 'w') as f:
            f.write(bundle)

        with open(filePath + 'templates/' + template_name_started, 'w') as f:
            f.write(bundleStarted)

        with open(filePath + 'function_store/' + function_name, 'w') as f:
            f.write(source_str)

        with open(filePath + 'templateDictionary.py', 'w') as f:
            f.write('tempDict = {}'.format(full_temp_dict))


def provWrap(func):
    '''
    Wrapper to decorate a function with in order to record provenance
    :param func: Target function object to wrap
    :return: An altered version of func which records provenance
    '''
    provWrap.count = 0

    def wrap(*args, **kwargs):
        name = func.__name__
        if name in praetor_settings.blacklist or praetor_rec != 0:
            return func(*args, **kwargs)
        try:
            module_loc = func.__module__
            if module_loc == '__main__':
                module_loc = 'main'
                module_name = ''
            else:
                module_name = '_' + module_loc
        except Exception:
            module_loc = 'undefined'
            module_name = ''

        try:
            input_list = [key for key in inspect.signature(func).parameters.keys()]
        except (TypeError, AttributeError, ValueError) as e:
            input_list = ['inputs_not_found']
            print('inputs for function {} not found'.format(name))

        longname = name
        template = name + module_name + '_template.provn'
        json_name = name + module_name
        template_name_started = name + module_name + 'started_template.provn'
        function_name = name + module_name + 'function.txt'
        inlist = list(args)
        filePath = praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/'
        jsonDir = filePath + 'json/'

        if not os.path.isfile(filePath + 'templates/' + template):
            create_template(func, template, template_name_started, function_name, filePath)

        try:
            defaultsKeys, defaultsVals = get_default_args(func)
        except TypeError:
            defaultsKeys, defaultsVals = [], []
        kwarg_list = []
        for key in input_list:
            if key in kwargs.keys():
                kwarg_list.append(kwargs[key])
            elif key in defaultsKeys:
                kwarg_list.append(defaultsVals[defaultsKeys.index(key)])

        inlist.extend(kwarg_list)


        if not praetor_settings.global_tracking:
            global_inputs = None
        else:
            global_start = get_globals()
            global_inputs = find_dict_differences(glob_vars[0], global_start)
            glob_vars.insert(0, global_start)

        startTime = datetimestamp()

        print('{} function_start: {}.{}'.format(praetor_settings.unique_string, module_loc, name))
        if praetor_settings.memory:
            try:
                mem_use, output = memory_usage((func, args, kwargs), retval=True)
                max_mem = str(max(mem_use))
            except:
                output = func(*args, **kwargs)
                max_mem = '0'
        else:
            output = func(*args, **kwargs)
            max_mem = '0'

        if not praetor_settings.global_tracking:
            global_outputs = None
        else:
            global_variables = get_globals()
            global_outputs = find_dict_differences(glob_vars[0], global_variables)
            glob_vars.insert(0, global_variables)

        print('{} function_end: {}.{}'.format(praetor_settings.unique_string, module_loc, name))
        endTime = datetimestamp()
        outlist = [output]
        provcall(inlist, outlist, template, longname, startTime, endTime, max_mem, json_name, jsonDir=jsonDir,
                 module_name=module_loc, global_outputs=global_outputs, global_inputs=global_inputs)
        call_func_list.append(inspect.stack()[1].function)
        return output

    return wrap


def provWrapOpen(func):
    '''
    Wrapper alter pythons built-in open function so that it records provenance
    :param func: The open function object (typically just open)
    :return: Altered functionality open which also records provenance
    '''
    def wrapOpen(*args, **kwargs):
        filePath = praetor_settings.provenance_directory + os.environ['PRAETOR_pipeline_id'] + '/'
        provDests = ['/json/', '/prov/', '/templates/', '/big_entities/', '/function_store/']
        provPaths = [filePath + x for x in provDests]
        fileName = args[0]

        try:
            gate = [True for x in provPaths if x in fileName]
        except TypeError:
            gate = [False]

        if any(gate):
            output = func(*args, **kwargs)

        else:
            longname = 'pythonBuiltinFileAccess'
            template = longname + '_template.prov'
            json_name = longname
            jsonDir = filePath + '/json/'
            startTime = datetimestamp()
            if praetor_settings.memory:
                try:
                    mem_use, output = memory_usage((func, args, kwargs), retval=True)
                    max_mem = str(max(mem_use))
                except:
                    output = func(*args, **kwargs)
                    max_mem = '0'
            else:
                output = func(*args, **kwargs)
                max_mem = '0'
            endTime = datetimestamp()
            mode = output.mode
            metaData = os.stat(fileName)
            inlist = [fileName, mode]
            outlist = []
            inlist.extend([x for i, x in enumerate(metaData) if i in [4, 6]])
            hashvalue = md5(args[0])
            inlist.append(hashvalue)
            provcall(inlist, outlist, template, longname, startTime, endTime, max_mem, json_name, jsonDir=jsonDir)
            call_func_list.append(inspect.stack()[0][3])
        return output

    return wrapOpen


def for_all_methods(decorator):
    '''
    Applies a target decorator to a class, was used to apply the provWrap but is currently out of use
    :param decorator: Decorator to add to a class
    :return: Class with altered functionality to record provenance
    '''
    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)):
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate

def for_all_methods_inherited(decorator):
    '''
    Applies a target decorator to a class, was used to apply the provWrap but is currently out of use
    :param decorator: Decorator to add to a class
    :return: Class with altered functionality to record provenance
    '''
    def decorate(cls):
        for attr in cls.__dict__:
            for name, fn in inspect.getmembers(cls, inspect.isroutine):
                setattr(cls, name, decorator(getattr(cls, fn)))
        return cls

    return decorate


def decorate_imported_function(func, decorator):
    '''
    Function to apply a decorator to an imported function
    :param func: Target function object
    :param decorator: Decorator to apply
    :return: Altered function which now records provenance
    '''
    func_type = type(func).__name__
    if isinstance(func, types.FunctionType):
        func = decorator(func)
    elif isinstance(func, types.BuiltinFunctionType):
        func = decorator(func)
    elif isinstance(func, types.ModuleType):
        func = decorate_all_in_module(func, decorator)
    else:
        print('function {} with type {} not supported by praetor and not wrapped'.format(func, func_type))
    return func


def decorate_all_in_module(module, decorator):
    '''
    Decorate the functions within imported modules
    :param module: Module to decorate
    :param decorator: decorator to apply
    :return: The function of the selected module is altered to record provenance
    '''
    mod_spec_blacklist = [x for x in praetor_settings.blacklist if module.__name__ in x]
    func_blacklist = [x[x.find('.')+1:] for x in mod_spec_blacklist]
    for name in dir(module):
        obj = getattr(module, name)
        try:
            obj_name = obj.__name__
        except:
            obj_name = obj
        if obj_name not in func_blacklist:
            if isinstance(obj, types.FunctionType):
                setattr(module, name, decorator(obj))
                # print(name, 'function')
            elif isinstance(obj, types.BuiltinFunctionType):
                setattr(module, name, decorator(obj))
                # print(name, 'builtin')
            elif isinstance(obj_name, types.ModuleType):
                # for_all_methods(decorator(obj_name))
                # print(name, 'module')
                decorate_all_in_module(obj, decorator)
            else:
                # print('passing')
                pass


